@@ -38,24 +38,31 @@ from ovirtcli.shell.summarycmdshell import SummaryCmdShell
 from ovirtcli.shell.filecmdshell import FileCmdShell
 from ovirtcli.shell.historycmdshell import HistoryCmdShell
 from ovirtcli.shell.infocmdshell import InfoCmdShell
-
-from ovirtcli.settings import OvirtCliSettings
-from ovirtcli.prompt import PromptMode
+from ovirtcli.shell.capabilitiescmdshell import CapabilitiesCmdShell
 
 from cli.command.help import HelpCommand
 from cli.messages import Messages
 
-from urlparse import urlparse
+from ovirtcli.events.event import Event
+from ovirtcli.listeners.errorlistener import ErrorListener
+from ovirtcli.infrastructure.settings import OvirtCliSettings
+from ovirtcli.shell.promptmode import PromptMode
+from cli.error import CommandError
 
+from ovirtcli.state.statemachine import StateMachine
+from ovirtcli.state.dfsastate import DFSAState
+from ovirtcli.shell.promptmanager import PromptManager
 
 class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
                   ShowCmdShell, ListCmdShell, UpdateCmdShell, \
                   RemoveCmdShell, AddCmdShell, DisconnectCmdShell, \
                   ConsoleCmdShell, PingCmdShell, StatusCmdShell, \
                   ClearCmdShell, FileCmdShell, HistoryCmdShell, \
-                  InfoCmdShell, SummaryCmdShell):
+                  InfoCmdShell, SummaryCmdShell, CapabilitiesCmdShell):
     OFF_LINE_CONTENT = [ConnectCmdShell.NAME, HelpCommand.name, 'exit', "EOF"]
+
     ############################# INIT #################################
+
     def __init__(self, context, parser, completekey='tab', stdin=None, stdout=None):
         cmd.Cmd.__init__(self, completekey=completekey, stdin=stdin, stdout=stdout)
         ConnectCmdShell.__init__(self, context, parser)
@@ -74,25 +81,41 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
         HistoryCmdShell.__init__(self, context, parser)
         InfoCmdShell.__init__(self, context, parser)
         SummaryCmdShell.__init__(self, context, parser)
+        CapabilitiesCmdShell.__init__(self, context, parser)
 
-        self.last_output = ''
+        self.__last_output = ''
         self.__input_buffer = ''
-        self.__org_prompt = ''
         self.__last_status = -1
 
-        self._set_prompt(mode=PromptMode.Disconnected)
+        self.onError = Event()  # triggered when error occurs
+        self.onInit = Event()  # triggered on init()
+        self.onExit = Event()  # triggered on exit
+        self.onPromptChange = Event()  # triggered onPromptChange
+        self.onSigInt = Event()  # triggered on SigInt fault
+
+        self.__prompt_manager = PromptManager(self)
+
+        self.__register_sys_listeners()
+        self.__register_dfsm_callbacks()
+        self.__init_promt()
+
         cmd.Cmd.doc_header = self.context.settings.get('ovirt-shell:commands')
         cmd.Cmd.undoc_header = self.context.settings.get('ovirt-shell:misc_commands')
-        cmd.Cmd.intro = OvirtCliSettings.INTRO
+        cmd.Cmd.intro = self.__get_intro()
 
         readline.set_completer_delims(' ')
-        signal.signal(signal.SIGINT, self.handler)
+        signal.signal(signal.SIGINT, self.__handler)
+
+        self.onInit.fire()
+
+    ############################ SHELL ##################################
 
-    ########################### SYSTEM #################################
     def cmdloop(self, intro=None, clear=True):
         try:
             if clear: self.do_clear('')
             return cmd.Cmd.cmdloop(self, intro)
+        except KeyboardInterrupt:
+            return self.cmdloop(intro="^C", clear=False)
         except Exception, e:
             self._error(str(e))
             return self.cmdloop(intro, clear=False)
@@ -115,61 +138,183 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
                     command.strip() not in EngineShell.OFF_LINE_CONTENT:
                 self._error(Messages.Error.INVALID_COMMAND % command)
             else:
-                if s.endswith('\\') and s != 'EOF':
-                    self._set_prompt(mode=PromptMode.Multiline)
-                    self.__input_buffer += ' ' + s.replace('\\', '') \
-                    if not s.startswith(' ') and self.__input_buffer != ''\
-                    else s.replace('\\', '')
-                    return
-                elif self.__input_buffer != '' and s != 'EOF':
-                    self.__input_buffer += ' ' + s \
-                    if not s.startswith(' ') else s
-                    s = self.__input_buffer
-                    self.__input_buffer = ''
-                    self._set_prompt(mode=PromptMode.Original)
+                try:
+                    if s.endswith('\\') and s != 'EOF':
+                        self.__set_prompt(mode=PromptMode.Multiline)
+                        self.__input_buffer += ' ' + s.replace('\\', '') \
+                        if not s.startswith(' ') and self.__input_buffer != ''\
+                        else s.replace('\\', '')
+                        return
+                    elif self.__input_buffer != '' and s != 'EOF':
+                        self.__input_buffer += ' ' + s \
+                        if not s.startswith(' ') else s
+                        s = self.__input_buffer
+                        self.__input_buffer = ''
+                        self.__set_prompt(mode=PromptMode.Original)
+                    return cmd.Cmd.onecmd(self, s)
+                finally:
+                    if self.context.status == \
+                       self.context.SYNTAX_ERROR \
+                       or \
+                       self.context.status == \
+                       self.context.COMMAND_ERROR \
+                       or \
+                       self.context.status == \
+                       self.context.UNKNOWN_ERROR:
+                        self.onError.fire()
+                    elif self.context.status == \
+                       self.context.COMMUNICATION_ERROR:
+                        self.__last_status = self.context.status
+                        self.onError.fire()
+                        StateMachine.communication_error()  # @UndefinedVariable
+                    elif self.context.status == \
+                       self.context.AUTHENTICATION_ERROR:
+                        self.__last_status = self.context.status
+                        self.onError.fire()
+                        StateMachine.unauthorized()  # @UndefinedVariable
+                    elif self.__last_status <> -1 and \
+                         (
+                            (
+                                self.__last_status == self.context.COMMUNICATION_ERROR
+                                and
+                                StateMachine.get_current_state() != DFSAState.COMMUNICATION_ERROR
+                            )
+                          or \
+                            (
+                                self.__last_status == self.context.AUTHENTICATION_ERROR
+                                and
+                                StateMachine.get_current_state() != DFSAState.UNAUTHORIZED
+                            )
+                         ):
+                        self.__set_prompt(
+                          mode=PromptMode.Original
+                        )
+                        self.__last_status = -1
 
-                result = cmd.Cmd.onecmd(self, s)
+    ########################### SYSTEM #################################
+    def __get_intro(self):
+        """
+        @return: the shell intro
+        """
+        return self.__produce_screen_adapted_message(
+                                      text=OvirtCliSettings.INTRO,
+                                      border="+"
+        )
 
-                # if communication error occurred, change prompt state
-                if self.context.status == self.context.COMMUNICATION_ERROR:
-                    self.__last_status = self.context.COMMUNICATION_ERROR
-                    self.owner._set_prompt(mode=PromptMode.Disconnected)
-                elif self.__last_status == self.context.COMMUNICATION_ERROR:
-                    self.__last_status = -1
-                    self.owner._set_prompt(mode=PromptMode.Original)
+    def __produce_screen_adapted_message(self, text, border, newline="\n\n"):
+        """
+        produces (pretty) screen width adapted message
+        
+        @param text: text to prin
+        @param border: border to use (default is '+')
+        @param newline: new line separator
+        """
+        return self.context.formatter.format_terminal(
+                          text=text,
+                          border=border,
+                          termwidth=self.context.terminal._get_width(),
+                          newline=newline
+        )
 
-                return result
+    def __init_promt(self):
+        self.__set_prompt(mode=PromptMode.Disconnected)
 
-    def _set_prompt(self, mode=PromptMode.Default):
-        if mode == PromptMode.Multiline:
-            if not self.__org_prompt:
-                self.__org_prompt = self.prompt
-            self.prompt = '> '
-        elif mode == PromptMode.Original:
-            if self.__org_prompt:
-                self.prompt = self.__org_prompt
-                self.__org_prompt = ''
-        elif mode == PromptMode.Disconnected or mode == PromptMode.Default:
-            if not self.__org_prompt and self.prompt != "(Cmd) ":
-                self.__org_prompt = self.prompt
-            self.prompt = self.context.settings.get('ovirt-shell:ps1.disconnected')
-        elif mode == PromptMode.Connected:
-            self.prompt = self.__get_connected_prompt()
+    def __register_dfsm_callbacks(self):
+        """
+        registers StateMachine events callbacks
+        """
+        StateMachine.add_callback(
+                      DFSAState.UNAUTHORIZED,
+                      self.__on_unauthorized_callback
+        )
+        StateMachine.add_callback(
+                      DFSAState.COMMUNICATION_ERROR,
+                      self.__on_communication_error_callback
+        )
+        StateMachine.add_callback(
+                      DFSAState.DISCONNECTED,
+                      self.__on_disconnected_callback
+        )
+        StateMachine.add_callback(
+                      DFSAState.CONNECTED,
+                      self.__on_connected_callback
+        )
+        StateMachine.add_callback(
+                      DFSAState.EXITING,
+                      self.__on_exiting_callback
+        )
 
-    def __get_connected_prompt(self):
-        if self.context.settings.get('ovirt-shell:extended_prompt'):
-            url = self.context.settings.get('ovirt-shell:url')
-            url_obj = urlparse(url)
-            if url_obj and hasattr(url_obj, 'hostname'):
-                return self.context.settings.get(
-                       'ovirt-shell:ps3.connected'
-                       ) % {
-                          'host':url_obj.hostname
-                       }
-        return self.context.settings.get('ovirt-shell:ps2.connected')
+    def __on_unauthorized_callback(self, **kwargs):
+        """
+        triggered when StateMachine.UNAUTHORIZED state is acquired
+        """
+        self.__set_prompt(
+            mode=PromptMode.Unauthorized
+        )
 
+    def __on_communication_error_callback(self, **kwargs):
+        """
+        triggered when StateMachine.COMMUNICATION_ERROR state is acquired
+        """
+        self.__set_prompt(
+            mode=PromptMode.Disconnected
+        )
 
-    def __persistCmdOptions(self, opts):
+    def __on_connected_callback(self, **kwargs):
+        """
+        triggered when StateMachine.CONNECTED state is acquired
+        """
+        self.context.history.enable()
+        if not self.context.settings.get('ovirt-shell:execute_command'):
+            self._print(
+               self.__produce_screen_adapted_message(
+                             text=OvirtCliSettings.CONNECTED_TEMPLATE,
+                             border='=',
+                             newline='\n'
+               ) % \
+               self.context.settings.get('ovirt-shell:version')
+            )
+
+        self.__set_prompt(
+            mode=PromptMode.Connected
+        )
+
+    def __on_disconnected_callback(self, **kwargs):
+        """
+        triggered when StateMachine.DISCONNECTED state is acquired
+        """
+        self.context.history.disable()
+        if not self.context.settings.get('ovirt-shell:execute_command'):
+            self._print(
+                "\n" + \
+                self.__produce_screen_adapted_message(
+                             text=OvirtCliSettings.DISCONNECTED_TEMPLATE,
+                             border='=',
+                             newline='\n'
+               )
+            )
+        self.__set_prompt(
+            mode=PromptMode.Disconnected
+        )
+
+    def __on_exiting_callback(self, **kwargs):
+        """
+        triggered when StateMachine.EXITING state is acquired
+        """
+        if StateMachine.get_origin_state() != DFSAState.DISCONNECTED \
+           and \
+           self.context.connection:
+            self.do_disconnect('')
+        else:
+            self.do_echo("\n")  # break shell prompt in to bash shell
+
+    def __register_sys_listeners(self):
+        self.onError += ErrorListener(self)
+
+    def __set_prompt(self, mode):
+        self.__prompt_manager.set_prompt(mode)
+
+    def __persist_cmd_options(self, opts):
         """
         Overrides config file options with cmdline's.
         """
@@ -181,18 +326,26 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
                     elif self.context.settings.has_key('cli:' + k):
                         self.context.settings['cli:' + k] = v
 
+    def _get_last_status(self):
+        return self.__last_status
 
     def onecmd_loop(self, s):
         opts, args = self.parser.parse_args()
         del args
-        self.__persistCmdOptions(opts)
+        self.__persist_cmd_options(opts)
         if opts.connect or self.context.settings.get('cli:autoconnect'):
-            self.do_clear('')
+            self.context._collect_connection_data()
+            if not self.context.settings.get('ovirt-shell:execute_command'):
+                self.do_clear('')
             self.do_connect(s)
             if opts.file:
-                cmd.Cmd.intro = None
                 self.do_file(opts.file)
-            self.cmdloop(clear=False)
+                self.do_exit('')
+            elif opts.execute_command:
+                self.__execute_command(opts.execute_command)
+                self.do_exit('')
+            else:
+                self.cmdloop(clear=False)
         else:
             self.cmdloop()
 
@@ -292,12 +445,38 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
             return None
 
     def _error(self, msg):
-        sys.stderr.write("\nerror: " + msg + "\n")
-        sys.stdout.write("\n")
+        # this is a custom error, consumer responsibility
+        # to filter out duplicate calls on event
+        self.onError.fire()
+        self.context._handle_exception(CommandError(msg))
 
-    def _print(self, msg):
-        sys.stdout.write("\n" + msg + "\n")
-        sys.stdout.write("\n")
+    def _print(self, msg, file=sys.stdout):  # @ReservedAssignment
+        """
+        prints message to the output device (default stdout).
+
+        @param msg: text to print
+        @param file: output device
+        """
+        self.context._pint_text(msg, file)
+
+    def __execute_command(self, execute):
+        """
+        executes command from the command line
+        via -E/--execute-command
+
+        @param execute: the command to execute
+        """
+        self.context.settings['cli:autopage'] = False
+        for command in execute.split(';'):
+            line = self.precmd(command)
+            self.print_line(line)
+            self.onecmd(line)
+
+    def __handler(self, signum, frame):
+        self.onSigInt.fire()
+        raise KeyboardInterrupt
+
+    ############################# COMMANDS #################################
 
     def do_EOF(self, line):
         """\
@@ -313,7 +492,8 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
 
         Ctrl+D
         """
-        self.emptyline(no_prompt=True)
+        self.onExit.fire()
+        StateMachine.exiting()  # @UndefinedVariable
         return True
 
     def do_exit(self, args):
@@ -330,7 +510,8 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
 
         exit
         """
-
+        self.onExit.fire()
+        StateMachine.exiting()  # @UndefinedVariable
         sys.exit(0)
 
     def do_help(self, args):
@@ -347,7 +528,6 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
 
         help show
         """
-
         if not args:
             self.context.execute_string('help\n')
         else:
@@ -359,7 +539,7 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
                     self.context.terminal.stdout.write('\n' + getattr(self, 'do_' + cmd).__doc__ + '\n')
                 else:
                     return self.context.execute_string('help ' + args + '\n')
-    ############################# SHELL #################################
+
     def do_shell(self, line):
         """\
         == Usage ==
@@ -378,10 +558,9 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
 
         ! ls -la
         """
-
         output = os.popen(line).read()
         print output
-        self.last_output = output
+        self.__last_output = output
 
     def do_echo(self, line):
         """\
@@ -401,13 +580,8 @@ class EngineShell(cmd.Cmd, ConnectCmdShell, ActionCmdShell, \
 
         echo str
         """
-
-        if self.last_output:
-            print line.replace('$out', self.last_output)
+        if self.__last_output:
+            print line.replace('$out', self.__last_output)
         elif line:
             print line
         else: print self.prompt
-    ############################## COMMON ################################
-    def handler(self, signum, frame):
-        self.emptyline(no_prompt=True)
-        sys.exit()
